"""Mixins for pydantic models"""
from __future__ import annotations

import abc
from bson import ObjectId
from motor import motor_asyncio
from pydantic import BaseModel
from pymongo.collection import Collection, ReturnDocument
from typing import TYPE_CHECKING, Any, List, Optional, Union, cast

from .db import get_db_manager
from .decoders.mongodb import AbstractMongoDBDecoder, BaseMongoDBDecoder
from .encoders.mongodb import AbstractMongoDBEncoder, BaseMongoDBEncoder
from .types import ObjectIdStr

if TYPE_CHECKING:
    from pydantic.typing import MappingIntStrAny  # isort: skip
    from pydantic.typing import AbstractSetIntStr, DictAny, DictIntStrAny, DictStrAny


class BaseDBMixin(BaseModel, abc.ABC):
    """Base class for Pydantic mixins"""

    id: Optional[ObjectIdStr] = None

    # Read-only (for public) field for store MongoDB id
    _doc: "DictAny" = {}

    # Encoders and decoders
    _mongodb_encoder: AbstractMongoDBEncoder = BaseMongoDBEncoder()
    _mongo_decoder: AbstractMongoDBDecoder = BaseMongoDBDecoder()

    class Config:
        allow_population_by_field_name = True
        json_encoders: "DictAny" = {ObjectId: lambda v: ObjectIdStr(v)}

    def __setattr__(self, key: Any, value: Any) -> Any:
        if key not in ["_doc"]:
            return super(BaseDBMixin, self).__setattr__(key, value)
        self.__dict__[key] = value
        return value

    @classmethod
    def _decode_mongo_documents(cls, document: "DictStrAny") -> "DictStrAny":
        """Decode and return MongoDB documents"""
        return cls._mongo_decoder(document)

    @classmethod
    def _encode_dict_to_mongo(cls, data: "DictStrAny") -> "DictStrAny":
        """Encode any dict to mongo query"""
        return cls._mongodb_encoder(data)

    def _encode_model_to_mongo(
        self,
        include: Union["AbstractSetIntStr", "DictIntStrAny"] = None,
        exclude: Union["AbstractSetIntStr", "DictIntStrAny"] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> DictStrAny:
        """Encode model to mongo query like pydantic.dict()"""
        model_as_dict = self.dict(
            include=include,
            exclude=exclude,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        return self._mongodb_encoder(model_as_dict)

    def dict(
        self,
        *,
        include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        by_alias: bool = False,
        skip_defaults: bool = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> DictStrAny:
        # Remove internal fields from serialized result
        if not exclude:
            exclude = {"_doc"}
        else:
            exclude = {"_doc", *exclude}

        return super(BaseDBMixin, self).dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

    def _update_model_from__doc(self) -> BaseDBMixin:
        """
        Update model fields from _doc dictionary
        (projection of a document from DB)
        """
        new_obj = self.parse_obj(self._doc)
        new_obj.id = self._doc.get("id")
        for k, field in new_obj.__fields__.items():
            field_default = getattr(field, "default", None)
            self.__dict__[k] = getattr(new_obj, k, field_default)
        return new_obj


class DBPydanticMixin(BaseDBMixin):
    """Help class for communicate of Pydantic model and MongoDB"""

    class Config:
        # DB
        collection: Optional[str] = None
        database: Optional[str] = None

    @classmethod
    async def get_collection(cls) -> Collection:
        db_name = getattr(cls.Config, "database", None)
        collection_name = getattr(cls.Config, "collection", None)
        if not db_name or not collection_name:
            raise ValueError("Collection or db_name is not configured in Config class")
        db_manager = get_db_manager()
        if not db_manager:
            raise RuntimeError("MongoDBManager not initialized")
        db = db_manager[db_name]
        if not db:
            raise ValueError('"%s" is not found in MongoDBManager.databases' % db_name)

        collection = db[collection_name]
        if not collection:
            collection = await db.create_collection(collection_name)
        return collection

    @staticmethod
    async def pre_save_validation(
        data: Union["DictAny", List["DictAny"]], many: bool = False
    ) -> Union["DictAny", List["DictAny"]]:
        """
        Async validation before save data to DB

        For use this method - override him in your model.
        You can also serializing data by returning a modified
        data dictionary (but I don't recommend).

        "many" flag will inform you about a mass operation.

        Called before:
            - create (with save)
            - update
            - save
            - bulk create
            - bulk update

        Usage example:

            class YourModel(DBPydanticMixin):
                other_model_id: ObjectIdStr

                async def validation(
                        data: Union["DictAny", List["DictAny"]], many: bool = False
                ) -> Union["DictAny", List["DictAny"]]:
                    other_model = await OtherModel.find_one({
                        '_id': data.get('other_model_id')
                    })
                    if not other_model:
                        raise TypeError(f'Other model with `{id}` id not found')
                    return data
        """
        return data

    @classmethod
    async def create(cls, fields: Union["DictAny", BaseModel]) -> DBPydanticMixin:
        """Create document by dict or pydantic model"""
        if isinstance(fields, BaseModel):
            fields = fields.dict(exclude_unset=True)
        document = cls.parse_obj(fields)
        await document.save()
        return document

    @classmethod
    async def count(cls, query: DictStrAny = None) -> int:
        """Return count by query or all documents in collection"""
        if not query:
            query = {}
        query = cls._encode_dict_to_mongo(query)
        collection = await cls.get_collection()
        return await collection.count_documents(query)

    @classmethod
    async def find_one(cls, query: DictStrAny) -> DBPydanticMixin:
        """Find and return model from db by pymongo query"""
        collection = await cls.get_collection()
        query = cls._encode_dict_to_mongo(query)
        result = await collection.find_one(query)
        if result:
            result = cls._decode_mongo_documents(result)
            model = cls.parse_obj(result)
            document_id = result.get("id")
            model._doc = result
            model.id = document_id
            return model
        return result

    @classmethod
    async def find_many(
        cls, query: "DictStrAny", return_cursor: bool = False
    ) -> Union[List[DBPydanticMixin], motor_asyncio.AsyncIOMotorCursor]:
        """
        Find documents by query and return list of model instances
        or query cursor
        """
        collection = await cls.get_collection()
        query = cls._encode_dict_to_mongo(query)
        cursor = collection.find(query)
        if return_cursor:
            return cursor

        documents = []
        async for _doc in cursor:
            _doc = cls._decode_mongo_documents(_doc)
            document = cls.parse_obj(_doc)
            document._doc = _doc
            documents.append(document)
        return documents

    @classmethod
    async def update_many(
        cls, query: "DictStrAny", fields: "DictAny", return_cursor: bool = False,
    ) -> Union[List[DBPydanticMixin], motor_asyncio.AsyncIOMotorCursor]:
        """
        Find and update documents by query
        """
        await cls.pre_save_validation(fields, many=True)
        collection = await cls.get_collection()
        query = cls._encode_dict_to_mongo(query)
        await collection.update_many(query, fields)
        return await cls.find_many(query, return_cursor)

    @classmethod
    async def bulk_create(
        cls, documents: Union[List[BaseModel], List["DictAny"]],
    ) -> List[DBPydanticMixin]:
        """Create many documents"""
        collection = await cls.get_collection()

        if not documents:
            return []

        if isinstance(documents[0], BaseModel):
            documents = [
                cls._encode_dict_to_mongo(d.dict())
                for d in cast(List[BaseModel], documents)  # noqa: types
            ]

        documents = cast(List["DictAny"], documents)  # noqa: types
        await cls.pre_save_validation(documents, many=True)

        result = await collection.insert_many(documents)
        inserted_ids = result.inserted_ids
        inserted_documents = []
        for i, document_id in enumerate(inserted_ids):
            document = cls.parse_obj(documents[i])
            document.id = document_id
            document._doc = cls._decode_mongo_documents(documents[i])
            inserted_documents.append(document)
        return inserted_documents

    async def reload(self) -> DBPydanticMixin:
        """Reload model data from MongoDB (get new document from db)"""
        collection = await self.get_collection()
        if not self.id:
            raise ValueError("Not found id in current model instance")
        _doc = await collection.find_one({"_id": self.id})
        if _doc:
            self._doc = self._decode_mongo_documents(_doc)
            self._update_model_from__doc()
        return self

    async def update(self, fields: Union[BaseModel, "DictAny"],) -> DBPydanticMixin:
        """
        Update Mongo document and pydantic instance.

        Parameters:
            - `fields`: updating fields (Pydantic model or dict)
        """
        if isinstance(fields, BaseModel):
            fields = fields.dict(exclude_unset=True)
        await self.pre_save_validation(fields)
        collection = await self.get_collection()
        if not self.id:
            raise ValueError("Not found id in current model instance")
        fields = self._encode_dict_to_mongo(fields)
        _doc = await collection.find_one_and_update(
            {"_id": self.id}, {"$set": fields}, return_document=ReturnDocument.AFTER
        )
        if _doc:
            self._doc.update(self._decode_mongo_documents(_doc))
            self._update_model_from__doc()
        return self

    async def save(self) -> DBPydanticMixin:
        collection = await self.get_collection()
        if not self.id:
            data = self._encode_model_to_mongo()
            await self.pre_save_validation(data)
            instance = await collection.insert_one(data)
            if instance:
                self.id = instance.inserted_id
                self._doc = {"id": self.id, **self.dict()}
        else:
            updated = {}
            data = self._encode_model_to_mongo(exclude={"id"})
            await self.pre_save_validation(data)
            for field, value in data.items():
                if self._doc.get(field) != value:
                    updated[field] = value
            if updated:
                instance = await collection.update_one(
                    {"_id": self.id}, {"$set": updated}
                )
                if instance:
                    self._doc.update(updated)
        return self

    async def delete(self) -> int:
        """Delete document from db"""
        collection = await self.get_collection()
        if not self.id:
            raise ValueError("Not found id in current model instance")
        result = await collection.delete_one({"_id": self.id})
        self._doc = {}
        return result.deleted_count



"""Encoders for MongoDB"""
from __future__ import annotations

import abc
from bson.decimal128 import Decimal128
from decimal import Decimal
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, List, Union, cast

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny


class AbstractMongoDBEncoder(abc.ABC):
    """Abstract MongoDB encoder"""

    @abc.abstractmethod
    def __call__(self, data: "DictStrAny") -> "DictStrAny":
        """Convert data from pydantic model or dict to dict supported MongoDB"""
        raise NotImplementedError()


def _recursive_iterator(
    data: Union["DictStrAny", List[Any]], transform_func: Callable[[Any], Any]
) -> Union["DictStrAny", List[Any]]:

    # Cast append func type
    append: Callable[[Union[str, int], Any], None]
    # Convert in list
    data = cast(List[Any], data)  # noqa
    _data = []
    iterator = enumerate(data)
    append = lambda k, v: _data.append(v)  # noqa: E731
    # Convert in dict
    if isinstance(data, dict):
        data = cast("DictStrAny", data)
        iterator = data.items()
        _data = {}
        append = lambda k, v: _data.update({k: v})  # noqa: E731
    # Iterate passed data
    for key, value in iterator:
        # Convert and replace value with transform function
        value = transform_func(value)
        # Recursive call if find sequence
        if isinstance(value, (list, dict)):
            value = _recursive_iterator(value, transform_func)
        # Update new data with update method (append for list and update for dict)
        append(key, value)
    # Return new data
    return _data


def _convert_enums(
    data: Union["DictStrAny", List[Any]]
) -> Union["DictStrAny", List[Any]]:
    """
    Convert Enum to Enum.value for mongo query

    Note: May be this solution not good
    """

    def enum_to_value(value: Union[Any, Enum]) -> Any:
        if isinstance(value, Enum):
            value = value.value
        return value

    return _recursive_iterator(data, enum_to_value)


def _convert_decimals(
    data: Union["DictStrAny", List[Any]]
) -> Union["DictStrAny", List[Any]]:
    """
    Convert decimal.Decimal to bson.decimal128.Decimal128
    """

    def python_decimal_to_bson_decimal(
        value: Union[Any, Decimal]
    ) -> Union[Any, Decimal128]:
        if isinstance(value, Decimal):
            value = Decimal128(value)
        return value

    return _recursive_iterator(data, python_decimal_to_bson_decimal)


class BaseMongoDBEncoder(AbstractMongoDBEncoder):
    """Base MongoDB encoder"""

    def __call__(self, data: "DictStrAny") -> "DictStrAny":
        data = cast("DictStrAny", _convert_enums(data))
        data = cast("DictStrAny", _convert_decimals(data))
        return data



"""Additional scheme pydantic validation errors"""
from pydantic.errors import PydanticValueError


class RangeBorderCrossing(PydanticValueError):
    code = "range.border_crossing"
    msg_template = "Make sure the borders do not cross"


class DatetimeBorderCrossing(RangeBorderCrossing):
    code = "datetime_range.border_crossing"



"""Types for pydantic models"""
from __future__ import annotations

from bson import ObjectId
from bson.errors import InvalidId
from datetime import datetime
from pydantic import BaseModel
from typing import TYPE_CHECKING, Any, Callable, Generator, List, Union

from .errors import DatetimeBorderCrossing

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny

    ValidatorClsMethod = Callable[[Any], Any]


class ObjectIdStr(str):
    """Field for validate string like ObjectId"""

    @classmethod
    def __get_validators__(cls) -> Generator["ValidatorClsMethod", None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, v: Union[ObjectId, str]) -> str:
        if isinstance(v, ObjectId):
            return str(v)
        else:
            try:
                ObjectId(str(v))
            except InvalidId:
                raise ValueError("Not a valid ObjectId")
            return v


class DateTimeRange(BaseModel):
    """
    Datetime range type.

    First elem - lower bound.
    Second elem - upper bound.
    """

    gte: datetime
    lte: datetime

    @classmethod
    def __get_validators__(cls) -> Generator["ValidatorClsMethod", None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, v: Union[List[Any], "DictStrAny"]) -> DateTimeRange:
        """Check border crossing"""
        gte, lte = None, None
        try:
            if isinstance(v, list):
                gte, lte = v
            elif isinstance(v, dict):
                gte = v.get("gte")
                lte = v.get("lte")
        except IndexError:
            return cls(gte=gte, lte=lte)

        if gte and lte:
            if gte > lte or lte < gte:
                raise DatetimeBorderCrossing()

        return cls(gte=gte, lte=lte)



import sys
from setuptools import setup, find_packages

from pydantic_odm import __version__

is_wheel = 'bdist_wheel' in sys.argv

excluded = []


with open("README.md", "r") as fh:
    long_description = fh.read()


def exclude_package(pkg):
    for exclude in excluded:
        if pkg.startswith(exclude):
            return True
    return False


def create_package_list(base_package):
    return ([base_package] +
            [base_package + '.' + pkg
             for pkg
             in find_packages(base_package)
             if not exclude_package(pkg)])


setup(
    # Metadata
    name='pydantic-odm',
    version=__version__,
    author='Ivan Galin',
    author_email='gin.voglgorad@gmail.com',
    url='https://github.com/i8enn/pydantic-odm/',
    download_url='',
    description='Small async ODM for MongoDB based on Motor and Pydantic',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX :: Linux',
        'Environment :: Console',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet',
    ],
    packages=create_package_list('pydantic_odm'),
    install_requires=[
        'pydantic>=1.1',
        'motor>=2.0',
    ],
    python_requires='>=3.6',
    zip_safe=False
)



"""Database connector module"""
from __future__ import annotations

from asyncio import AbstractEventLoop, get_running_loop
from motor import motor_asyncio
from threading import Lock
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny

    DatabaseSettingsType = Dict[str, Dict[str, Any]]


class MongoDBManagerMeta(type):
    """MongoDBManager metaclass for implement singleton behavior"""

    # Main mongodb manager instance (for singleton implementation)
    _instance: Optional[MongoDBManager] = None

    # Lock for thread-save
    _lock: Lock = Lock()

    def __call__(
        cls, *args: Tuple[Any], **kwargs: DictStrAny
    ) -> Optional[MongoDBManager]:
        with cls._lock:
            if not cls._instance:
                cls._instance = super().__call__(*args, **kwargs)
        return cls._instance


class MongoDBManager(metaclass=MongoDBManagerMeta):
    """
    Singleton class for create main interface for working on many Mongo
    database connections from one place.

    Setting signature::
        {
            'connection_alias': {
                'SETTING_NAME': 'SETTINGS_VALUE'
            }
        }

    Very minimal connection configuration (Not recommended)::

        'minimal': {}

    .. note:: connection to 'minimal' Mongo database on mongodb://localhost:27017

    Minimal connection configuration::

        'minimal': {
            'NAME': 'mongo-db-name'
        }

    Full connection configuration::

        'full_configuration': {
            'NAME': 'mongo-db-name',
            'HOST': 'mongodb://localhost',
            'PORT': 27017,
            'USERNAME': 'mongo_user',
            'PASSWORD': 'mongo_password',
            'AUTHENTICATION_SOURCE': 'admin',
            'AUTH-MECHANISM': 'SCRAM-SHA-256'
        }

    Connection params description:
        - `NAME`: (str) Mongo database name for connection. If not setup - using configuration alias.
        - `HOST`: (str) Mongo hostname. Configured by default as in PyMongo (`localhost`).
        - `PORT`: (int) Mongo port. Configured by default as in PyMongo (27017).
        - `USERNAME`: (str) Mongo administration username. Default - empty.
        - `PASSWORD`: (str) Mongo administration password. Default - empty.
        - `AUTHENTICATION_SOURCE`: (str) Mongo database name for authentication
           source. Configured by default as in PyMongo ('admin')
        - `AUTH-MECHANISM`: (str) Mongo authentication mechanism. If not setup -
          PyMongo automatically selects a mechanism depending on the version of
          MongoDB (See https://api.mongodb.com/python/current/examples/authentication.html#default-authentication-mechanism)


    Usage::

    On startup application execute initial connect.

        def startup():
            MongoDBManager.initial_connections({
                'default': {
                    'NAME': 'mongo-db-name',
                    'HOST': 'mongodb://localhost',
                    'PORT': 27017,
                    'USERNAME': 'mongo_user',
                    'PASSWORD': 'mongo_password',
                    'AUTHENTICATION_SOURCE': 'admin',
                    'AUTH-MECHANISM': 'SCRAM-SHA-256'
                },
                'minimal_configured_db': {
                    'NAME': 'mongo-db-name'
                }
            })

    Get collection in all place in your code:
    from pydantic_odm.db import MongoDBManager
    ...
    any_collections = MongoDBManager.default.any_collections
    """  # noqa: E501

    # Event loop for passing to MotorClient
    _loop: AbstractEventLoop

    # Passed settings
    settings: DatabaseSettingsType
    # Created connections
    connections: Dict[str, motor_asyncio.AsyncIOMotorClient] = {}
    # Configured databases
    databases: Dict[str, motor_asyncio.AsyncIOMotorDatabase] = {}
    # Init database flag
    is_init: bool = False

    def __init__(
        self, database_settings: DatabaseSettingsType, loop: AbstractEventLoop = None
    ):
        self.settings = database_settings or {}
        if not loop:
            loop = get_running_loop()
        self._loop = loop

    def __getitem__(self, item: str) -> Optional[motor_asyncio.AsyncIOMotorDatabase]:
        return self.databases.get(item, None)

    async def init_connections(self) -> MongoDBManager:
        """Create connections to Mongo databases"""
        if self.is_init:
            return self

        if not self.settings:
            raise RuntimeError("Not found database configurations in MongoDBManager")

        for alias, configuration in self.settings.items():
            connection_params = {
                "username": configuration.get("USERNAME"),
                "password": configuration.get("PASSWORD"),
                "host": configuration.get("HOST"),
                "port": configuration.get("PORT"),
                "authSource": configuration.get("AUTH_SOURCE", ""),
            }
            if "OPTIONAL_PARAMETERS" in configuration:
                connection_params.update(**configuration["OPTIONAL_PARAMETERS"])
            auth_mech = configuration.get("AUTH_MECHANISM")
            if auth_mech:
                connection_params["authMechanism"] = auth_mech
            connection_params["io_loop"] = self._loop
            client = motor_asyncio.AsyncIOMotorClient(**connection_params)
            db_name = configuration.get("NAME", alias)
            self.connections[alias] = client
            db = client[db_name]
            self.databases[alias] = db

        self.is_init = True

        return self


def get_db_manager() -> Optional[MongoDBManager]:
    """Return initialized singleton mongodb manager"""
    return MongoDBManager._instance



"""MongoDB decoders"""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny


class AbstractMongoDBDecoder(abc.ABC):
    """Abstract MongoDB decoder"""

    @abc.abstractmethod
    def __call__(self, data: "DictStrAny") -> "DictStrAny":
        """Main mongodb encoder func"""
        raise NotImplementedError()


class BaseMongoDBDecoder(AbstractMongoDBDecoder):
    """Base MongoDB decoder"""

    def __call__(self, data: "DictStrAny") -> "DictStrAny":
        """
        Decode mongodb document.

        Rename field `_id` to `id`.
        """
        data = data.copy()
        data.pop("id", None)
        document_id = data.pop("_id", None)
        decoded_data = {"id": document_id}
        for k, v in data.items():
            if isinstance(v, list):
                v_list = []
                for item in v:
                    if isinstance(item, dict):
                        item = self.__call__(item)
                    v_list.append(item)
                decoded_data[k] = v_list
            elif isinstance(v, dict):
                decoded_data[k] = self.__call__(v)
            else:
                decoded_data[k] = v
        return decoded_data
